from enum import Enum
from typing import Callable, Dict, Optional, Any
from uuid import UUID, uuid4
from abc import ABC, abstractmethod

from pydantic import BaseModel, Field

MSG_HS_NODE = "MSG_HS_NODE"
MSG_HS_ACK = "MSG_HS_ACK"
MSG_TASK_REQ = "MSG_TASK_REQ"
MSG_TASK_FIN = "MSG_TASK_FIN"
MSG_STEP_FIN = "MSG_STEP_FIN"
MSG_HEARTBEAT = "MSG_HEARTBEAT"


class TaskStatus(str, Enum):
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


class WorkflowRunner(ABC):
    @abstractmethod
    def kickoff(self, inputs: Dict[str, str]) -> None:
        pass

    @abstractmethod
    def output(self) -> str:
        pass

    @abstractmethod
    def ongoing_steps(self) -> list:
        pass


class Workflow(BaseModel):
    name: str
    description: str
    inputs: Dict[str, str]
    step_description: list[str]
    workflowRunner: type = Field(exclude=True)


class AgenticConfig(BaseModel):
    id: Optional[UUID] = None
    name: str
    workflows: Dict[str, Workflow] = {}


class Task(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    inputs: Dict[str, str]
    status: TaskStatus = Field()
    output: str | None

class WrongFolderError(Exception):
    pass


class AgenticMessage(BaseModel):
    type: str


class AgenticHandshakeMessage(AgenticMessage):
    type: str = MSG_HS_NODE
    node: str


class TaskFinishedMessage(AgenticMessage):
    type: str = MSG_TASK_FIN
    task_id: str
    status: TaskStatus
    result: str | None


class StepFinishedMessage(AgenticMessage):
    type: str = MSG_STEP_FIN
    task_id: str
    step: int
    result: str


class TaskRequest(BaseModel):
    workflow: str
    inputs: Dict[str, str]
    task_id: UUID = Field(default_factory=uuid4)
    node_id: str


class AgenticTaskRequestMessage(BaseModel):
    type: str = MSG_TASK_REQ
    task: TaskRequest



from agenticos.node.models import AgenticConfig, Workflow, WorkflowRunner

workflows : dict[Workflow] = {}

# Example workflow
# Import the Crew class. If you used the flow from CrewAI docs the following import should work
from {{folder_name}}.crew import {{class_name}}Crew as Crew

# Implement runner for the workflow. This will be instantiated every time the workflow is run.
class CrewAIRunner(WorkflowRunner):
    def __init__(self):
        super().__init__()
        self.crew = Crew()

    # Pass the kickoff function of the crew.
    def kickoff(self, inputs: dict[str, str]):
        self.crew.crew().kickoff(inputs)

    # Final output of the workflow
    def output(self):
        return (self.crew.crew().tasks[-1].output.raw)

    # Intermediate steps of the workflow
    def ongoing_steps(self):
        return self.crew.crew().tasks

# Define the workflow name
my_workflow_name = "research_and_report"
# Define the workflow
workflows[my_workflow_name] = Workflow(
    # Name of the workflow
    name=my_workflow_name,
    # Description of the workflow
    description="This workflow will get you a joke which is not a PR nightmare",
    # Inputs of the workflow. This is what you pass to Crew.kickoff function
    inputs={"topic": "The topic to joke about"},
    # Description of the steps in the workflow
    step_description=[t.description for t in Crew().crew().tasks],
    # Pass the runner for the workflow
    workflowRunner=CrewAIRunner,
)

config = AgenticConfig(name="Test Node", workflows=workflows)



import logging
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread

log = logging.getLogger(__name__)

class HTTPHealthEndpoint(BaseHTTPRequestHandler):
    def do_GET(self):
        if self.path == "/health":
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(b'{"status": "ok"}')
        else:
            self.send_response(404)
            self.end_headers()
            self.wfile.write(b"Not found")

class HTTPHealthServer(Thread):
    def __init__(self, port: int):
        super().__init__()
        self.port = port

    def run(self):
        self.server = HTTPServer(("localhost", self.port), HTTPHealthEndpoint)
        log.info(f"Starting health server on port {self.port}")
        self.server.serve_forever()
        log.info("Health server stopped")


    def stop(self):
        log.debug("Health server stop requested")
        self.server.shutdown()



import logging

from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    AUTH_TOKEN: str = ""
    REGISTRY_URL: str = "ws://localhost:8080"
    NODE_MODE: str = "httpserver"
    HTTP_PORT: int = 8000
    AGENTIC_CONFIG_PATH: str = "src/agentic/agentic_config.py"
    HTTP_HEALTHCHECK : bool = False
    LOG_LEVEL: str = "INFO"

    
settings = Settings()

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

_nameToLevel = {
    'CRITICAL': CRITICAL,
    'FATAL': FATAL,
    'ERROR': ERROR,
    'WARN': WARNING,
    'WARNING': WARNING,
    'INFO': INFO,
    'DEBUG': DEBUG,
    'NOTSET': NOTSET,
}

_env_level = settings.LOG_LEVEL
_log_level = logging.INFO
if _env_level not in _nameToLevel:
    print(f"Invalid log level: {_env_level}. Defaulting to INFO")
else:
    _log_level = _nameToLevel[_env_level]
print("Setting log level to: ", _log_level) 

logging.basicConfig(
    level=_log_level,
    format="%(asctime)s | %(name)s | %(levelname)s | %(message)s",
)


import click

from ..node.settings import settings
from .init_agentic import init_agentic
from .run_agentic import HTTPSERVER, REGISTRY, run_agentic


@click.group()
def agenticos():
    """Top-level command group for agenticos."""


@click.command()
@click.option("-b", "--backend", type=click.Choice(["crewai"]), default="crewai")
@click.option("-f", "--force", is_flag=True, help="Overwrite existing files")
def init(backend, force):
    """Initialize a new AgenticOS project. Has to be run in the root folder of your project."""
    click.echo(
        "Initializing a new AgenticOS project... with backend: {}".format(backend)
    )
    init_agentic(backend, force)


@click.command()
@click.option(
    "-M", "--mode", type=click.Choice([HTTPSERVER, REGISTRY]), default=settings.NODE_MODE
)
@click.option("-P", "--port", type=int, default=settings.HTTP_PORT)
@click.option("-R", "--registry", type=str, default=settings.REGISTRY_URL)
@click.option("--dev", is_flag=True, help="Run in development mode")
def run(mode, port, registry, dev):
    """Run agentic node."""
    click.echo("Running agentic in {} mode".format(mode))
    run_agentic(mode=mode, port=port, dev=dev, registry=registry)


agenticos.add_command(init)
agenticos.add_command(run)



from uuid import UUID

from dotenv import load_dotenv
from fastapi import BackgroundTasks, FastAPI, HTTPException
from fastapi.responses import JSONResponse

from agenticos.node.models import Task, TaskStatus, Workflow

from .agentic_config import workflows

load_dotenv()

app = FastAPI()

tasks: dict[UUID, Task] = {}


def run_task(workflow: Workflow, task: Task) -> None:
    try:
        workflow.kickoff_function(task.inputs)
        task.output = workflow.output_function()
        task.status = TaskStatus.COMPLETED
    except Exception as e:
        task.output = str(e)
        task.status = TaskStatus.FAILED


@app.get(
    "/node/description",
    summary="Get the description of the node",
    response_model=dict[str,Workflow],
)
def description() -> dict[str,Workflow]:
    return workflows


@app.post("/workflow/{workflow_name}/run")
async def run(workflow_name : str, inputs: dict[str, str], background_tasks: BackgroundTasks) -> str:
    if workflow_name not in workflows:
        raise HTTPException(status_code=404, detail="Workflow not found")
    task = Task(inputs=inputs, status=TaskStatus.RUNNING, output="")
    tasks[task.id] = task
    background_tasks.add_task(run_task, workflows[workflow_name], task)
    return str(task.id)


@app.get("/task/{task_id}")
def get_task(task_id: str) -> Task:
    if UUID(task_id) not in tasks:
        raise HTTPException(status_code=404, detail="Task not found")
    return tasks[UUID(task_id)]


@app.get("/tasks")
def get_tasks() -> list[str]:
    return [str(tk) for tk in tasks.keys()]


@app.get("/health")
def health() -> JSONResponse:
    return JSONResponse({"status": "ok"})
